/*
   dtp.c

   Implementation of Data Transfer Process

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (shesmax@ru.hilti.com)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: dtp.c,v 1.21 2000/08/15 17:31:54 shesmax Exp $ */

#include <sys/types.h>

#include <unistd.h>     /* for close */
#include <fcntl.h>      /* for O_ macros */
#include <stdio.h>      /* for sscanf */
#include <stdlib.h>     /* for strtoul */
#include <ctype.h>      /* for toupper */
#include <string.h>     /* for strlen */

#include <sys/socket.h>    /* socket related stuff */
#include <netinet/ip.h>
#include <arpa/inet.h>     /* for inet_ntoa */

#include "libra.h"      /* for UNUSED_PARAM */
#include "pi.h"         /* for sinPeerCtrl */
#include "parser.h"     /* for GetPath */
#include "retcodes.h"
#include "dtp.h"
#include "dataopen.h"
#include "chkport.h"
#include "sockets.h"    /* for INVALID_SOCKET */
#include "xbuf.h"

/* Socket related */
struct sockaddr_in sinPeerData;     /* Peer */
struct sockaddr_in sinSrvrData;     /* Our host */
int hDataSocket = INVALID_SOCKET;   /* data connection */

/* Connection parameters */
int bPassive;                       /* 1 = passive connection */
static char cTransferMode;
static char cType;
static char cStructure;
static char cFormat;
static char nByteSize;

int nMaxMap;               /* Maximum mapping length;
                              should be positive int because of socket
                              interface parameter types. */
int nAsciiBufSize =
               128 * 1024; /* no more than MAX_INT/2, see SendAscii();
                              128 KB now */

nb_t nDtPassed;            /* current tranfer position */
nb_t nDtTotal;             /* total file size for RETR */
nb_t nRestartAt;           /* restart position */

static int bDataTransferInProgress = 0;
static int bAbortTransfer = 0;

static int bChecksums = RC_FALSE;
static int bPassiveOnly = RC_FALSE;

/*
* DTP housekeeping
*/

/*
   Initializes the state of the connection.
   All parameters retrieve default values.
   It is assumed that PI is already initialized.
*/
int InitDTP()
{
   /* default user data port is the same as its control port */
   sinPeerData = sinPeerCtrl;

   bPassive = 0;
   cType = TYPE_ASCII;
   cFormat = FORMAT_NONPRINT;
   cStructure = STRUCTURE_FILE;
   cTransferMode  = TRANSFER_MODE_STREAM;

   nMaxMap = getpagesize();
   nMaxMap = (10L * 1024L * 1024L / nMaxMap) * nMaxMap;  /* approx. 10 MB */

   nRestartAt = 0;
   bDataTransferInProgress = 0;

   return 1;
}

/* Frees resources allocated for the connection state. */
void ShutdownDTP()
{
   if( hDataSocket != INVALID_SOCKET )
      close( hDataSocket );
}

/*
* FTP command handlers
*/

/* Allows to specify a non-default data port for the user side. */
int Port( char* args )
{
   int h1, h2, h3, h4;
   int p1, p2;
   int nChars = 0;
   struct sockaddr_in addr;

   if( PassiveOnlyMode() )
      return SendReply( "502 Command not implemented\r\n" );
     
   /* Parse the arguments and return error code if something is
      ill-formed. */
   if
   (
      sscanf
      (
         args, "%d,%d,%d,%d,%d,%d%n",
         &h1, &h2, &h3, &h4,
         &p1, &p2,
         &nChars
      ) != 6 || nChars == 0 ||
      args[ nChars ] ||
      !IS_PORT_NUMBER( h1 ) || !IS_PORT_NUMBER( h2 ) ||
      !IS_PORT_NUMBER( h3 ) || !IS_PORT_NUMBER( h4 ) ||
      !IS_PORT_NUMBER( p1 ) || !IS_PORT_NUMBER( p2 )
   )
      return SendReply( "501 Syntax error in port name\r\n" );

   /* The port specification was OK, construct peer's address. */
   addr.sin_addr.s_addr = htonl
   (
      (h1 << 24) | (h2 << 16) | (h3 << 8) | h4
   );
   addr.sin_port = htons( (p1 << 8) | p2 );
   addr.sin_family = AF_INET;
   if( !CheckPortAddr( &sinPeerCtrl, &addr ) )
      return SendReply( "501 Syntax error in port name\r\n" );
   sinPeerData = addr;
   bPassive = 0;
   return SendReply( "200 PORT command OK\r\n" );
}

/* Request to be a passive endpoint. */
int Pasv( char* args )
{
   long addr;
   int h1, h2, h3, h4;
   short port;
   int p1, p2;

   /* create data socket and force it to listen for incoming connections */
   if( ExtraParameters( args ) )
      return SendReplies();
   if( CreateDataSocket( &hDataSocket) != RC_OK )
   {
      PError( 425 ); /* see comments for OpenDataConnection() */
      return SendReplies();
   }
   if( listen( hDataSocket, 1 ) < 0 )
   {
      close( hDataSocket );
      return SendReply( "425 Can\'t open passive data connection\r\n" );
   }
   bPassive = 1;

   /* Build the standard reply */
   addr = ntohl( sinSrvrData.sin_addr.s_addr );
   h1 = (addr >> 24) & 0xFF;
   h2 = (addr >> 16) & 0xFF;
   h3 = (addr >> 8) & 0xFF;
   h4 = addr & 0xFF;
   port = ntohs( sinSrvrData.sin_port );
   p1 = (port >> 8) & 0xFF;
   p2 = port & 0xFF;

   return SendReplyF
   ( 
      "227 Entering passive mode (%d,%d,%d,%d,%d,%d)\r\n", 
      h1, h2, h3, h4, p1, p2 
   );
}

/*
   Sets the data representation types.
   Only ascii, image and local 8 are supported.
*/
int Type( char* args )
{
   char buf[ 16 ];

   if( !*args )
      return SendReply( "501 Need type specifier\r\n" );

   switch( toupper( *args ) )
   {
   case TYPE_ASCII:
      switch( args[ 1 ] )
      {
      case ' ':
         if
         (
            (tolower( args[ 2 ] ) != 'n' && tolower( args[ 2 ] ) != 't') ||
            args[ 3 ]
         )
            goto __error__;
         break;
      case '\0':
         break;
      default:
         goto __error__;
      }
      cType = TYPE_ASCII;
      cFormat = FORMAT_NONPRINT;
      break;
   case TYPE_IMAGE:
      if( args[ 1 ] )
         goto __error__;
      cType = TYPE_IMAGE;
      cFormat = ' '; /* for display purposes only */
      break;
   case TYPE_LOCAL:
      if( args[ 1 ] != ' ' || args[ 2 ] != '8' || args[ 3 ] )
         goto __error__;
      cType = TYPE_LOCAL;
      cFormat = '8'; /* for display purposes only */
      nByteSize = 8;
      break;
   case TYPE_EBCDIC:
      goto __error__;
   }
   snprintf( buf, sizeof buf, "200 Type: %c %c\r\n", cType, cFormat );
   return SendReply( buf );

__error__:
   return SendReply( "504 Supported are: TYPE A [N|T], I and L 8\r\n" );
}

/* Specifies the file structure */
int Stru( char* args )
{
   if( args[ 1 ] )
      return SendReply( "501 Syntax error in parameter\r\n" );
   if( toupper( *args ) != STRUCTURE_FILE )
      return SendReply( "504 Only STRU F is supported\r\n" );
   cStructure = STRUCTURE_FILE;
   return SendReply( "200 Structure is File\r\n" );
}

/* Specifies the transfer mode */
int Mode( char* args )
{
   if( args[ 1 ] )
      return SendReply( "501 Syntax error in parameter\r\n" );
   if( toupper( *args ) != TRANSFER_MODE_STREAM )
      return SendReply( "504 Only MODE S is supported\r\n" );
   cTransferMode = TRANSFER_MODE_STREAM;
   return SendReply( "200 Current transfer mode is Stream" );
}

/* Sends the given file to the remote host */
int Retr( char* args )
{
   char* szPath;

   if( !GetPath( &args, &szPath, GP_OBLIGATORY ) )
     return SendReplies();
   return DoRetrieve( &hDataSocket, szPath );
}

/*
   Accept the data sent by the user DTP and store it as the file with
   the given name. If such file exists, overwrite it.
*/
int Stor( char* args )
{
   char* szPath;

   if( !GetPath( &args, &szPath, GP_OBLIGATORY ) )
     return SendReplies();
   return DoStore( &hDataSocket, szPath, nRestartAt ? 0 : O_TRUNC );
}

/*
   Stores the data sent by the remote user under the name unique in the
   current directory.
*/
int Stou( char* args UNUSED_PARAM )
{
   /* RFC 959 specifies that STOU has no parameters, but the
      Berkeley's FTP client sends the name as a parameter to the
      command. In accordance to the rule "Be liberal in what you
      accept ... " I decided to discard that parameter and work
      as if there was no extra input after the command. */
   int result;
   char name[] = "./libraXXXXXX";

   if( mktemp( name ) != NULL )
      result = DoStore( &hDataSocket, name, O_EXCL ); /* do not allow to
                                                         overwrite an existing
                                                         file */
   else {
      PError( 450 );
      result = RC_ERROR;
   }
   return result;
}

/* Appends the data sent by the remote user at the end of the given file. */
int Appe( char* args )
{
   char* szPath;

   if( !GetPath( &args, &szPath, GP_OBLIGATORY ) )
      return SendReplies();
   return DoStore( &hDataSocket, szPath, O_APPEND );
}

/* Reserve storage at the server side. */
int Allo( char* args UNUSED_PARAM )
{
   return SendReply
   (
      "202 Command not implemented, superfluous at this site.\r\n"
   );
}

int Rest( char* args )
{
   char* endp;
   nb_t value;
   unsigned long arg;

   if( !*args )
      return SendReply( "501 Need a stream offset\r\n" );

   arg = strtoul( args, &endp, 10 );
   value = arg;
   if( *endp || (off_t)value < 0 || arg != value ) /* type independent check */
      return SendReply( "501 Illegal number\r\n" );
   nRestartAt = value;
   return SendReply( "350 Ok, send file transfer request\r\n" );
}

/*
   Query and control DTP
*/

char* GetRepresentationType()
{
   switch( cType )
   {
   case TYPE_ASCII:
      return "Ascii";
   case TYPE_IMAGE:
      return "Image";
   default:
      return "Local 8";
   }
}

char* GetFileStructure()
{
   return "File";
}

char* GetTransferMode()
{
   return "Stream";
}

char* GetDataConnType()
{
   return bPassive ? "Passive" : "Active";
}

char* GetDataConnAddr()
{
   if( hDataSocket != INVALID_SOCKET )
   {
      /* static buffer is necessary to avoid the interference with other
         calls to inet_ntoa. */
      static char buf[ 30 ];
      
      snprintf
      (
         buf, sizeof buf, "%s port %d",
         inet_ntoa( sinPeerData.sin_addr ),
         ntohs( sinPeerData.sin_port )
      );
      return buf;
   }
   return "closed";
}

int DataTransferInProgress()
{
   return bDataTransferInProgress;
}

void StartTransfer()
{
   bDataTransferInProgress = RC_TRUE;
}

void StopTransfer()
{
   bDataTransferInProgress = RC_FALSE;
}

int AbortTransfer()
{
   return bAbortTransfer;
}

void PostTransferAbort()
{
   bAbortTransfer = RC_TRUE;
}

void CancelTransferAbort()
{
   bAbortTransfer = RC_FALSE;
}

nb_t DataToSend()
{
   return nDtTotal;
}

nb_t DataSent()
{
   return nDtPassed;
}

int ComputeChecksums()
{
   return bChecksums;
}

void SetChecksumMode( int on )
{
   bChecksums = on;
}

int PassiveOnlyMode()
{
   return bPassiveOnly;
}

void SetPassiveOnlyMode()
{
   bPassiveOnly = RC_TRUE;
}
