/* 
   pi.c

   Protocol Interprenter.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (shesmax@ru.hilti.com)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: pi.c,v 1.21 2000/08/15 17:31:57 shesmax Exp $ */

#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/utsname.h>

#include <netinet/in.h> /* Internet related stuff */
#include <netinet/ip.h>
#include <arpa/inet.h>  /* for inet_ntoa */
#include <arpa/telnet.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>

#include "libra.h"
#include "pi.h"
#include "dtp.h"
#include "parser.h"
#include "retcodes.h"
#include "listdir.h"
#include "mlsx.h"       /* MlsxListFacts */
#include "sockets.h"
#include "log.h"
#include "xbuf.h"

static const char sVersion[] = "Libra FTP daemon"
#ifdef LIBRA_VERSION
" " LIBRA_VERSION
#endif
;

/* Requests & Replies */
static XBuf xbReply;    /* module static data - used in several functions */
static XBuf xbCmd;      /* declared static to reduce number of alloc/dealloc */
static XBuf xbCmdAccum; /* command accumulation buffer - has to be static */
static XBuf xbOob;      /* declared static to reduce number of alloc/dealloc */
static XBuf xbOobAccum; /* OOB cmd accumulation buffer - has to be static */

/* Addresses */
typedef struct sockaddr* PSaddr;    /* shorter name */
struct sockaddr_in sinPeerCtrl;
struct sockaddr_in sinSrvrCtrl;

/*
* Out-of-band data
*/

static char* ReadCmdOob( void );
static void OobData()   /* SIGURG handler */
{
   char* sCmd;

   if( !DataTransferInProgress() )
      return;  /* Do not bother */
   
   if( (sCmd = ReadCmdOob()) == 0 )
      return;  /* The control connection is closed */

   if( strcasecmp( sCmd, "ABOR" ) == 0 )
      PostTransferAbort();
   else if( strcasecmp( sCmd, "STAT" ) == 0 )
      SendReplyF( "213 Sent %ld of %ld bytes\r\n", DataSent(), DataToSend() );
}

/*********************************************************************
*  Initailization/Deinitialization
*/

#define INITIAL_SIZE 160      /* the value doesn't matter */
static int PIInitBuffers( void )
{
   int n;

   n = 0;
   do {
      if (CreateFixedXBuf( &xbReply, INITIAL_SIZE ) == XBUF_OVERFLOW)
         break;
      n++;
      if (CreateExpandableXBuf( &xbCmd, INITIAL_SIZE ) == XBUF_OVERFLOW)
         break;
      n++;
      if (CreateExpandableXBuf( &xbCmdAccum, INITIAL_SIZE ) == XBUF_OVERFLOW)
         break;
      n++;
      if (CreateExpandableXBuf( &xbOob, INITIAL_SIZE ) == XBUF_OVERFLOW)
         break;
      n++;
      if (CreateExpandableXBuf( &xbOobAccum, INITIAL_SIZE ) == XBUF_OVERFLOW)
         break;
      n++;
   } while( 0 );

   switch( n )
   {
      case 5:
         return 1;
      case 4:
         DestroyXBuf( &xbOob );
         /* fall through */
      case 3:
         DestroyXBuf( &xbCmdAccum );
         /* fall through */
      case 2:
         DestroyXBuf( &xbCmd );
         /* fall through */
      case 1:
         DestroyXBuf( &xbReply );
         /* fall through */
      default:
         return 0;
   }
}

static void PIDestroyBuffers( void )
{
   DestroyXBuf( &xbReply );
   DestroyXBuf( &xbCmd );
   DestroyXBuf( &xbCmdAccum );
   DestroyXBuf( &xbOob );
   DestroyXBuf( &xbOobAccum );
}

static int InitSigHandlers( void )
{
   struct sigaction sa;

   /* Out-of-band data */
   sa.sa_handler = &OobData;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = SA_RESTART;
   if( sigaction( SIGURG, &sa, 0 ) == -1 )
      LogEvent
      ( 
         LLOG_WARN, "Failed to setup handler for out-of-band data: %s\n",
         strerror( errno )
      );
   return 1;
}

/*
* Setup the contol connection.
*/

static int CtrlSocketInit( void )
{
   int x;

   /* Get the address of the control connection. We need it
      because we should open the data connection on the same
      ip address.
    */
   x = sizeof( sinSrvrCtrl );
   /* sinSrvrCtrl is 0 for the case getsockname fails */
   if( getsockname( STDIN_FILENO, (PSaddr)&sinSrvrCtrl, &x ) == -1 )
      LogEvent( LLOG_WARN, "getsockname: %s\n", strerror( errno ) );
   /* ignore x */

   x = sizeof( sinPeerCtrl );
   if( getpeername( STDIN_FILENO, (PSaddr)&sinPeerCtrl, &x ) == -1 )
      LogEvent( LLOG_WARN, "getpeername: %s\n", strerror( errno ) );
   /* ignore x */

   if
   (
      /* Timeout handling */
      (x = fcntl( STDIN_FILENO, F_GETFL )) == -1 ||
      fcntl( STDIN_FILENO, F_SETFL, x | O_NONBLOCK ) == -1 ||
      (x = fcntl( STDOUT_FILENO, F_GETFL )) == -1 ||
      fcntl( STDOUT_FILENO, F_SETFL, x | O_NONBLOCK ) == -1
   )
   {
      LogEvent( LLOG_WARN, "fcntl(F_[GS]ETFL): %s\n", strerror( errno ) );
      return 0;
   }

   /* To receive the SIGURG signal */
   if( fcntl( STDOUT_FILENO, F_SETOWN, getpid() ) == -1 )
      LogEvent( LLOG_WARN, "fcntl(2, F_SETOWN): %s\n", strerror( errno ) );

   /* IP option "Better latency" for control connection */
   x = IPTOS_LOWDELAY;
   if
   ( 
      setsockopt( STDIN_FILENO, IPPROTO_IP, IP_TOS, &x, sizeof x ) == -1 ||
      (
         x = IPTOS_LOWDELAY, 
         setsockopt( STDOUT_FILENO, IPPROTO_IP, IP_TOS, &x, sizeof x ) == -1 
      )
   )
      LogEvent( LLOG_WARN, "IPTOS_LOWDELAY: %s\n", strerror( errno ) );

   x = 1;
   if( setsockopt( STDIN_FILENO, SOL_SOCKET, SO_OOBINLINE, &x, sizeof x ) < 0 )
      LogEvent
      ( 
         LLOG_WARN, "Failed to setup inline out-of-band data: %s\n",
         strerror( errno )
      );

   return 1;
}

/*
* Initialize PI module.
* Return TRUE on success, FALSE on fail.
*/
int InitPI()
{
   static char szErr[] = "421 Cannot initialize the control connection\r\n";
   int buffersInitialized;

   buffersInitialized = 0;
   do {
      if (!PIInitBuffers())
         break;
      buffersInitialized = 1;
      if (!InitSigHandlers())
         break;
      if (!CtrlSocketInit())
         break;
      return 1;
   } while( 0 );

   if( buffersInitialized )
      PIDestroyBuffers();

   /* Send the refusal to the peer but without any buffering */
   write( STDOUT_FILENO, szErr, sizeof szErr - 1 );
   return 0;
}

/*
* Exit routine for protocol interprenter.
* Called only if InitPI has succeeded.
*/
void ShutdownPI()
{
   PIDestroyBuffers();
}

/*********************************************************************
   Read a command from the remote client. 
   
   RFC 1123 requirements:
   -  The command can be of any size.
   -  There can be more than one command in the received TCP segment.
   -  FTP should abandon any Telnet options negotiation.
*/

/* Process Telnet commands */
static void TelnetProcess( XBuf* xbuf )
{
   unsigned char* p;
   unsigned char* e;
   unsigned char* s;

   /* 
      Set e -> \r. We put \0 there intentionaly - this is used in algorithm 
      of parsing Telnet commands.
   */
   e = xbuf->data + xbuf->nUsed - 1;   
   *e = '\0';   
#if 1
   p = s = xbuf->data;
      
   for( p = s = xbuf->data; p < e; ++p )
   {
      if( *p == IAC )   /* Interpret As Command */
      {
         switch( p[ 1 ] )
         {
         case DO:
            if( p + 2 < e )
            {
               /* Answer to remote: IAC WONT <option code> */
               p[ 1 ] = WONT;
               write( STDOUT_FILENO, p, 3 );
            }
            /* else violation of Telnet IAC DO \0 */
            p += 2;
            continue;

         case WILL:
            if( p + 2 < e )
            {
               /* Answer to remote: IAC DONT <option code> */
               p[ 1 ] = DONT; 
               write( STDOUT_FILENO, p, 3 );
            }
            /* else violation of Telnet IAC WILL \0 */
            p += 2;
            continue;

         /* 
            I am not shure we can get following two commands because they
            are answers to option negotiation, what we won't initiate.
         */
         case DONT:  
         case WONT: 
            p += 2;  /* skip three characters in command stream */
            continue;

         case IAC:   /* Escaped IAC */
            ++p;     /* copy IAC into the command */
            break;
         default: 
            /* 
               Here we can have in p[ 1 ] either one of the remaining
               Telnet commands (NOP, DM, IP, AYT etc) or some kind of
               violation of Telnet stream - IAC followed by some 
               "normal" character. I would prefer to skip this sequence.
            */
            ++p;
            continue;
         }
      }

      /* *p is some character of command request */
      if( s != p )
         *s = *p;
      ++s;
   }
   *s = '\0';
#endif
}

/*
* Check if the buffer contains line terminator (\r\n) and returns the line
* length including the terminator.
* If no line terminator is found return 0.
*/
static int CheckLineEnd( XBuf* xb )
{
   char* p;

   for( p = xb->data; ; )
   {
      p = memchr( p, '\r', xb->data + xb->nUsed - p );
      if( !p )
         break;
      p++; /* go over \r */
      if( p < xb->nUsed + xb->data && *p == '\n' )
         return p + 1 - xb->data;
   }
   return 0;
}

static int GetCommand( XBuf* dst, XBuf* src, int len )
{
   /* copy the command and \r to the command buffer */
   ClearXBuf( dst );
   if( WriteXBuf( dst, src->data, len - 1 ) == XBUF_OVERFLOW )
      return 0;

   /* remove the head of the accumulating buffer */
   memmove( src->data, src->data + len, src->nUsed - len );
   src->nUsed -= len;

   return 1;
}

static char* ReadCmd( void )
{
   char buf[ 4096 ];
   int nBytes, lineLen;
  
   /* Read the command either from the save buffer or from the socket */
   while( ( lineLen = CheckLineEnd( &xbCmdAccum ) ) == 0 )
   {
      nBytes = ReadSocket( STDIN_FILENO, buf, sizeof buf );
      if (nBytes <= 0)
         return 0;
      if( WriteXBuf( &xbCmdAccum, buf, nBytes ) == XBUF_OVERFLOW )
         return 0;
   }
   if( !GetCommand( &xbCmd, &xbCmdAccum, lineLen ) )
      return 0;

   TelnetProcess( &xbCmd );

   LogEvent( LLOG_CMD, "Cmd: %s\n", xbCmd.data );
   return xbCmd.data;
}

static char* ReadCmdOob( void )
{
   /* FIXME */
   return ReadCmd();
}

/*********************************************************************
   Upper-level PI functions.
*/

static int Welcome()
{
   char* nodename;
   struct utsname uts;
   char* buf;

   /* 
      Get the name of the node either from environment (for virtual
      hosts) or system wide.
    */
   nodename = getenv( "TCPLOCALHOST" ); /* tcp-server specific */
   if( !nodename )
      if( uname( &uts ) != -1 )
         nodename = uts.nodename;
   if( !nodename )
      nodename = "";

   buf = (char*)malloc( strlen( nodename ) + strlen( sVersion ) + 4 );
   if( buf )
   {
      int r;
      
      sprintf( buf, "%s (%s)", nodename, sVersion );
      r = DirInfo( 220, buf );
      free( buf );
      return r;
   }
   return SendReply( "220 Welcome!\r\n" );
}

/* Communicates with peer by reading its requests and executing them. */
int RunPI()
{
   if( Backend() || Welcome() == RC_OK )
      for(;;)
      {
         char* sCmd;

         if( (sCmd = ReadCmd()) == 0 || ExecCommand( sCmd ) != RC_OK )
            break;
      }
   return RC_OK;
}

/*
* Command handlers
*/

int Abor( char* unused UNUSED_PARAM )
{
   return SendReply( "226 No open data connection\r\n" );
}

/* Returns the type of the operating system used at the server's site. */
int Syst( char* args )
{
   if( ExtraParameters( args ) )
      return SendReplies();
   return SendReply( "215 UNIX (L8)\r\n" );
}

/* Auxiliary command for STAT * */
static int LongListEntry( const char* name, const struct stat* st, void* nFnd )
{
   char sPrefix[ 57 ];

   if( ++(*(int*)nFnd) == 1 )
      AccumReply( "212-\r\n" );
   LsLongPrefix( st, sPrefix, sizeof sPrefix );
   AccumReply( sPrefix );
   AccumReplyBuf( name, strlen( name ) ); 
   return AccumReply( "\r\n" );     
}

char* PeerName()
{
   char* szPeerName;
   char* s;
   
   /* Try to detect the name from tcp-server first */
   s = getenv( "TCPREMOTEHOST" );
   if( !s )
   {
      if( (szPeerName = (char*)malloc( 16 )) != 0 )
      {
         strncpy( szPeerName, inet_ntoa( sinPeerCtrl.sin_addr ), 16 );
         szPeerName[ 15 ] = '\0';
      }
   }
   else
   {
      size_t n = strlen( s ) + 1;
      
      if( (szPeerName = malloc( n )) != 0 )
         memcpy( szPeerName, s, n );
   }
   return szPeerName;
}

/* Status response */
int Stat( char* args )
{
   static char sOn[] = "On";
   static char sOff[] = "Off";
  
   if( !*args )      /* Return the information about the server */
   {
      char* sPeer = PeerName();

      AccumReplyF
      (
         "211- Server status:\r\n"
         " Connected to: %s\r\n"
         " Type: %s, Structure: %s, Mode: %s,\r\n"
         " Checksums: %s, Passive data connections only: %s\r\n"
         " Data connection: %s (%s)\r\n"
         " Command buffer: %d bytes, Reply buffer: %d bytes\r\n"
         "211 \r\n",
         sPeer,
         GetRepresentationType(), GetFileStructure(), GetTransferMode(),
         ComputeChecksums() ? sOn : sOff, PassiveOnlyMode() ? sOn : sOff,
         GetDataConnType(), GetDataConnAddr(),
         xbCmd.nSize, xbReply.nSize
      );
      free( sPeer );
   }
   else     /* STAT * */
   {
      int nEntries = 0;
		
      if( Ls( args, LongListEntry, (void*)&nEntries ) == RC_OK )
         AccumReplyF( "212 %d entries\r\n", nEntries );
      else
         PError( 212 );
   }
   return SendReplies();
}

/* Sends OK reply to the peer. */
int Noop( char* unused UNUSED_PARAM )
{
   return SendReply( "200 OK\r\n" );
}

/*
* RFC 2389 - Feature negotiation mechanism for the File Transfer Protocol
*/

/* 
   Lists non-standard commands supported at this moment.
   The extensions listed here are not documented now, hence
   it is unknown how to return them. Temporary we return their
   names only. It is also undefined whether FEAT should return
   SITE commands. We list them here too.
*/
int Feat( char* args )
{
   if( !ExtraParameters( args ) )
   {
      AccumReply
      (
         "211-Extensions supported\r\n"
         " MDTM\r\n"
         " REST STREAM\r\n"
         " SIZE\r\n"
         " TVFS\r\n"
      );
      
      AccumReply( " MLST" );
      MlsxListFacts( MLF_ALL_FACTS );
      AccumReply( "\r\n" );

      AccumReply
      (
         " SITE CHECKSUM\r\n"
         " SITE CHMOD\r\n"
         " SITE IDLE\r\n"
         " SITE MINFO\r\n"
         " SITE NEWER\r\n"
         " SITE UMASK\r\n"
         "211 END\r\n"
      );
   }
   return SendReplies();
}

/* Allows to set server specific options. */  
int Opts( char* args )
{
   static char szCmd[] = "MLST";
  
   if( strncasecmp( args, szCmd, sizeof szCmd - 1 ) == 0 )
      return MlsxOpts( args + sizeof szCmd - 1 );
   return SendReply( "501 Invalid syntax\r\n" );
}

/*
* Replies to the peer
*/

/* 
   Last Reply Status - holds the status of the serie of ...Reply... calls. 
   This simplifies the code, because it avoids the necessity to check the
   return code of ...Reply... functions all the time. You write the program as
   if they all can't fail and check the return code only of the last one.
*/
static int nLRS = RC_OK;

/* 
   This flag is necessary for the implementation of reply audit. 
   We can't write a reply from the server to the audit log at the time
   of buffer flush because probably we'll do it several times for a long reply.
   It would be very unpleasant for the reader to find one part of the 
   answer at line number X and the rest of this answer several hundred
   lines below (on heavy used server ;-).
   Hence the program writes only the start of an answer to the audit log.
*/
static int bStartOfReply = 1;

static int FlushReplyBuf( XBuf* buf, int socket )
{
   if( bStartOfReply )
   {
      LogEvent( LLOG_REPLY, "%.*s", buf->nUsed, buf->data );
      bStartOfReply = 0;
   }
   return FlushSockXBuf( buf, socket );
}

int AccumReplyBuf( const char* text, size_t sz )
{
   if( nLRS == RC_OK )
      nLRS = PutSockXBuf( &xbReply, text, sz, FlushReplyBuf, STDOUT_FILENO );
   return nLRS;
}

static int PutReply( const char* fmt, va_list args )
{
   XBuf tmp;

   if( CreateExpandableXBuf( &tmp, 256 ) == XBUF_OVERFLOW )
      return nLRS = RC_ERROR;
   if( VPrintfXBuf( &tmp, fmt, args ) == XBUF_OVERFLOW )
      nLRS = RC_ERROR;
   else
      AccumReplyBuf( tmp.data, tmp.nUsed );
   DestroyXBuf( &tmp );
   return nLRS;
}

int AccumReplyF( const char* fmt, ... )
{
   va_list args;

   va_start( args, fmt );
   PutReply( fmt, args );
   va_end( args );
   return nLRS;
}

int SendReplyBuf( const char* text, size_t nBytes )
{
   AccumReplyBuf( text, nBytes );
   return nLRS == RC_OK ? SendReplies() : nLRS;
}

int SendReplyF( const char* fmt, ... )
{
   va_list args;

   va_start( args, fmt );
   PutReply( fmt, args );
   va_end( args );
   return nLRS == RC_OK ? SendReplies() : nLRS;
}

int SendReplies()
{
   if( bStartOfReply )
      LogEvent( LLOG_REPLY, "%.*s", xbReply.nUsed, xbReply.data );
   if( nLRS == RC_OK )
      nLRS = FlushSockXBuf( &xbReply, STDOUT_FILENO );
   ClearXBuf( &xbReply );
   bStartOfReply = 1;
   return nLRS;
}

void PError( int nCode )
{
   AccumReplyF( "%d %s\r\n", nCode, strerror( errno ) );
}
