/*
   sockets.c

   Wrappers for socket calls.

   Sockets are opened in non-blocking mode.
   Current implementation uses select() for timeouts.
   Timeouts are configurable by redefining of DEFAULT_TIMEOUT macro.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (shesmax@ru.hilti.com)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: sockets.c,v 1.10 2000/08/15 17:31:57 shesmax Exp $ */

#include <unistd.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/stat.h>      /* for struct stat */

#include "sockets.h"
#include "retcodes.h"
#include "dtp.h"           /* OOB data handler queries and controls DTP */

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0x4000
#endif

/*
* Timeouts
*/

static unsigned long nTimeout = DEFAULT_TIMEOUT;

unsigned long GetSocketTimeout()
{
   return nTimeout;
}

int SetSocketTimeout( unsigned long nTimeout_ )
{
   if( nTimeout_ > DEFAULT_TIMEOUT )
      return RC_FALSE;
   nTimeout = nTimeout_;
   return RC_TRUE;
}

/*
* Working with socket
*/

/* Waits for the data to be available for reading on socket nSocket */
static int Wait4DataAvail( int nSocket )
{
   fd_set fds;
   struct timeval tv = {nTimeout, 0};

   FD_ZERO( &fds );
   FD_SET( nSocket, &fds );
   return select( nSocket + 1, &fds, 0, 0, &tv ) == 1;
}

/* 
   Waits for the socket nSocket to be available for writing.
   
   Returns
   
   RC_OK          socket is available for writing
   RC_TIMEDOUT    timeout expired
   RC_ERROR       error on the socket.
*/
static int WaitOk2Write( int nSocket )
{
   fd_set fds;
   struct timeval tv = {nTimeout, 0};

   FD_ZERO( &fds );
   FD_SET( nSocket, &fds );
   if( select( nSocket + 1, 0, &fds, 0, &tv ) != -1 )
      return FD_ISSET( nSocket, &fds ) ? RC_OK : RC_TIMEDOUT;
   return RC_ERROR;
}

/* Sends the data to the remote host through the socket */
int WriteSocket( int nSocket, const void* pBuf, int nCount )
{
   struct msghdr mh;
   struct iovec iov;
   struct stat statbuf;
   int is_socket;
   int r;

   if( fstat( nSocket, &statbuf ) == -1)
      return RC_ERROR;
   if( S_ISFIFO( statbuf.st_mode ) )
      return RC_ERROR; /* we can't handle SIGPIPE for them */
   is_socket = 0;
   if( S_ISSOCK( statbuf.st_mode ) )
      is_socket = 1;

   iov.iov_base = (void *)pBuf;
   iov.iov_len = nCount;
   mh.msg_name = NULL;
   mh.msg_namelen = 0;
   mh.msg_iov = &iov;
   mh.msg_iovlen = 1;
   mh.msg_control = NULL;
   mh.msg_controllen = 0;
   mh.msg_flags = 0;
   while( iov.iov_len )
   {
      int nWritten;

      if( AbortTransfer() )
      {
         CancelTransferAbort();
         return RC_ABORTED;
      }
      
      r = WaitOk2Write( nSocket );
      if (r == RC_OK)
      {
         nWritten = is_socket
                     ? sendmsg( nSocket, &mh, MSG_NOSIGNAL )
                     : writev( nSocket, &iov, 1 );
         if( nWritten != -1 )
         {
            iov.iov_len -= nWritten;
            *((char **)&iov.iov_base) += nWritten;
         }
         else
            r = RC_ERROR;
      }
      switch( r )
      {
      case RC_TIMEDOUT:
         return RC_TIMEDOUT;
      case RC_ERROR:
         if( errno == EINTR )
            continue;
         return RC_ERROR;
      }
   }

   return RC_OK;
}

/* Reads not more than nCount bytes from the socket nSocket into the buffer */
int ReadSocket( int nSocket, void* pBuf, int nCount )
{
   if( AbortTransfer() )
   {
      CancelTransferAbort();
      return RC_ABORTED;
   }
   if( !Wait4DataAvail( nSocket ) )
      return RC_TIMEDOUT;
   return read( nSocket, pBuf, nCount );
}

/* Connects socket with the given address addr. */
int ConnectSocket( int nSocket, struct sockaddr_in* addr )
{
   if( connect( nSocket, (struct sockaddr*)addr, sizeof *addr ) != -1 )
      return RC_OK;
   if( errno != EINPROGRESS )
      return RC_ERROR;
   return WaitOk2Write( nSocket );
}

/* Accepts incoming connection requests on the given socket */
int AcceptSocket( int nSocket, struct sockaddr_in* addr )
{
   int x = sizeof *addr;

   if( !Wait4DataAvail( nSocket ) )
      return INVALID_SOCKET;
   return accept( nSocket, (struct sockaddr*)addr, &x );
}
