/*
   xbuf.c

   XBuf allows to create either a buffer of fixed length or an expandable
   buffer. The size of the expandable buffer is limited only be the
   size of the heap.

   The implementation provides several functions that allow to append
   the desired text at the end of the buffer.

   This buffer is heavily used in the program for different purposes.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (shesmax@ru.hilti.com)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: xbuf.c,v 1.10 2000/08/15 17:31:58 shesmax Exp $ */

#include <stdio.h>
#include <stdarg.h>
#include <malloc.h>
#include <string.h>

#include "xbuf.h"
#include "retcodes.h"

#include "sockets.h"

/* Auxiliary function for CreateFixedXBuf and CreateExpandableXBuf */
static int Create_( XBuf* pBuf, size_t nSize )
{
	if( (pBuf->data = (char*)malloc( nSize )) == 0 )
		return XBUF_OVERFLOW;
	pBuf->nUsed = 0;
	pBuf->nSize = nSize;
	return XBUF_OK;
}

int CreateFixedXBuf( XBuf* pBuf, size_t nSize )
{
   if( !pBuf )
      return XBUF_BADPARAM;
   pBuf->bFixed = 1;
   return Create_( pBuf, nSize );
}

int CreateExpandableXBuf( XBuf* pBuf, size_t nSize )
{
   if( !pBuf )
      return XBUF_BADPARAM;
   pBuf->bFixed = 0;
   return Create_( pBuf, nSize );
}

void DestroyXBuf( XBuf* pBuf )
{
   if( pBuf )
      free( pBuf->data );
}

void ClearXBuf( XBuf* pBuf )
{
   if( pBuf )
      pBuf->nUsed = 0;
}

int WriteXBuf( XBuf* pBuf, const char* data, size_t nBytes )
{
   size_t nSize;

   if( !pBuf || !data )
      return XBUF_BADPARAM;
   if( (nSize = pBuf->nUsed + nBytes) > pBuf->nSize )
   	if( pBuf->bFixed )
      {
         /* Reduce the size of the portion that should be copied */
         if( (nBytes -= nSize - pBuf->nSize) == 0 )
            return XBUF_OVERFLOW;
      }
      else
      {
			/* grow the buffer */
         size_t nAlloc;
   		char* p;

         for( nAlloc = pBuf->nSize << 1; nAlloc < nSize; nAlloc <<= 1 )
				;
         if( (p = realloc( pBuf->data, nAlloc )) == 0 )
   	   	return XBUF_OVERFLOW;
	   	pBuf->data = p;
         pBuf->nSize = nAlloc;
      }

	memcpy( pBuf->data + pBuf->nUsed, data, nBytes );
	pBuf->nUsed += nBytes;
	return nBytes;
}

int VPrintfXBuf( XBuf* pBuf, const char* fmt, va_list args )
{
   if( !pBuf )
      return XBUF_BADPARAM;
   for(;;)
	{
		char* buf = pBuf->data + pBuf->nUsed;
		int nSize = pBuf->nSize - pBuf->nUsed;
		int nBytes;

		if( nSize > 0)
		{
         nBytes = vsnprintf( buf, nSize, fmt, args );
         if (nBytes != -1 && nBytes <= nSize) {
            pBuf->nUsed += nBytes;
            return nBytes;
         }
		}

      /* We should increase the size of the buffer */

      /* Fixed buffer can't grow hence deny the request */
      if( pBuf->bFixed )
         return XBUF_OVERFLOW;

		nSize = pBuf->nSize << 1;
		if( (buf = realloc( pBuf->data, nSize )) == 0 )
         return XBUF_OVERFLOW;
		pBuf->data = buf;
		pBuf->nSize = nSize;
	}
}

int PrintfXBuf( XBuf* pBuf, const char* fmt, ... )
{
	va_list args;
	int result;

	va_start( args, fmt );
	result = VPrintfXBuf( pBuf, fmt, args );
	va_end( args );
	return result;
}

/*
* Working with sockets 
*/

int FlushSockXBuf( XBuf* pXBuf, int hSocket )
{
   int nResult = WriteSocket( hSocket, pXBuf->data, pXBuf->nUsed );

   if( nResult == RC_OK )
      ClearXBuf( pXBuf );
   return nResult;
}

int PutSockXBuf
( 
   XBuf* pXBuf, const char* buf, int nSize, 
   int (*Flush)( XBuf*, int ), int hSocket 
)
{
   int nBytes;

   while( nSize > 0 )
   {
      if( (nBytes = WriteXBuf( pXBuf, buf, nSize )) != XBUF_OVERFLOW )
      {
         nSize -= nBytes;
         buf += nBytes;
      }
      else if( (nBytes = (*Flush)( pXBuf, hSocket )) != RC_OK )
         return nBytes;
   }
   return RC_OK;
}
